"""
Geometry, topology and density measures
"""
import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.neighbors import KNeighborsRegressor

def l3(X, y, normalize=True):
    """
    Calculates the nom-linearity of a linear regressor (L3) measure.

    Linearly interpolates both input (X) and output (y) values of each pair of samples with similar output values. Generated l=n-1 synthetic samples and then measures the mean squared error of a linear regressor, fitted with original data and evaluated on synthetic points. By default performs a normalization of samples.
    
    .. math::

        L3=\\frac{1}{l}\sum_{i=1}^{l}(f(x'_i) - y'_i)^2

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels

    :rtype: float
    :returns: L3 score
    """

    X = np.copy(X)
    y = np.copy(y)

    if normalize:
        for f_id in range(X.shape[1]):
            X[:,f_id] -= np.min(X[:,f_id])
            X[:,f_id] /= np.max(X[:,f_id])
        y -= np.min(y)
        y /=np.max(y)

    sorted_y_idx = np.argsort(y)

    a_idx = sorted_y_idx[:-1]
    b_idx = sorted_y_idx[1:]
    pairs = np.concatenate((a_idx[:, np.newaxis], b_idx[:, np.newaxis]), axis=1)
    rand = np.random.random(pairs.shape[0]).reshape(-1,1)

    #interpolate input X
    distance = X[pairs[:,0]]-X[pairs[:,1]]
    rand_distance = rand * distance
    X_new = X[pairs[:,1]] + rand_distance

    #interpolate output y
    distance = y[pairs[:,0]]-y[pairs[:,1]]
    rand_distance = rand[:,0] * distance
    y_new = y[pairs[:,1]] + rand_distance

    err = LinearRegression().fit(X,y).predict(X_new) - y_new
    return np.sum(np.power(err, 2))/X_new.shape[0]


def s4(X, y, normalize=True):
    """
    Calculates the non-linearity of a nearest neighbor regressor (S4) measure.

    Linearly interpolates both input (X) and output (y) values of each pair of samples with similar output values. Generated l=n-1 synthetic samples and then measures the mean squared error of a nearest neighbor regessor, fitted with original data and evaluated on synthetic points. By default performs a normalization of samples.
    
    .. math::

        S4=\\frac{1}{l}\sum_{i=1}^{l}(NN(x'_i) - y'_i)^2

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels

    :rtype: float
    :returns: S4 score
    """

    X = np.copy(X)
    y = np.copy(y)

    if normalize:
        for f_id in range(X.shape[1]):
            X[:,f_id] -= np.min(X[:,f_id])
            X[:,f_id] /= np.max(X[:,f_id])
        y -= np.min(y)
        y /=np.max(y)

    sorted_y_idx = np.argsort(y)

    a_idx = sorted_y_idx[:-1]
    b_idx = sorted_y_idx[1:]
    pairs = np.concatenate((a_idx[:, np.newaxis], b_idx[:, np.newaxis]), axis=1)
    rand = np.random.random(pairs.shape[0]).reshape(-1,1)

    #interpolate input X
    distance = X[pairs[:,0]]-X[pairs[:,1]]
    rand_distance = rand * distance
    X_new = X[pairs[:,1]] + rand_distance

    #interpolate output y
    distance = y[pairs[:,0]]-y[pairs[:,1]]
    rand_distance = rand[:,0] * distance
    y_new = y[pairs[:,1]] + rand_distance

    err = KNeighborsRegressor().fit(X,y).predict(X_new) - y_new
    return np.sum(np.power(err, 2))/X_new.shape[0]
    
def t2(X, y):
    """
    Calculates the average number of examples per dimension (T2) measure.

    Returns number of samples per number of features. Higher values indicate simpler problems.
        
    .. math::

        T2=\\frac{n}{d}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels

    :rtype: float
    :returns: T2 score
    """
    return X.shape[0]/X.shape[1]


"""Classification tests."""
import problexity as px
from sklearn.datasets import make_classification
import numpy as np

np.random.seed(123)

def _get_comparison(metric):
    reps = 20
    res = np.zeros((reps, 2))
    rs = np.random.randint(100,10000,reps)

    for r in range(reps):
        X_simple, y_simple = make_classification(n_samples=500, n_features=5, n_redundant=0, 
                        n_informative=5, n_clusters_per_class=2, n_classes=2, flip_y=0, 
                        class_sep=10, weights=[0.5, 0.5], random_state=rs[r])

        X_complex, y_complex = make_classification(n_samples=500, n_features=10, n_redundant=0, 
                        n_informative=10, n_clusters_per_class=2, n_classes=2, flip_y=0.2, 
                        class_sep=0.01, weights=[0.9, 0.1], random_state=rs[r])

        r_s = metric(X_simple, y_simple)
        r_c = metric(X_complex, y_complex)

        res[r, 0] = r_s
        res[r, 1] = r_c

    compare = np.mean(res[:,0]<res[:,1])
    return compare
 
### F1 Tests ###

def test_F1():
    metric = px.classification.f1
    assert _get_comparison(metric)>0.5

def test_F1_close_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.f1(X,y)
    assert value<0.0001

def test_F1_equal_to_one():
    X, y = make_classification(random_state=115)
    X = np.concatenate((X[y==0],X[y==0]), axis=0)
    y = np.concatenate((y[y==0],np.ones((len(y[y==0])))))
    value = px.classification.f1(X,y)
    assert value==1

def test_F1_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.f1(X,y)
    assert 0 <= value <= 1 

def test_F1_single_class():
    X, y = make_classification(random_state=115, weights=[1., .0], flip_y=0.)
    value = px.classification.f1(X,y)
    assert np.isnan(value)

def test_F1_emptyset():
    X, y = [], []
    value = px.classification.f1(X,y)
    assert np.isnan(value)

### F1v Tests ###

def test_F1v():
    metric = px.classification.f1v
    assert _get_comparison(metric)>0.5

def test_F1v_close_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.f1v(X,y)
    assert value<0.0001

def test_F1v_equal_to_one():
    X, y = make_classification(random_state=115)
    X = np.concatenate((X[y==0],X[y==0]), axis=0)
    y = np.concatenate((y[y==0],np.ones((len(y[y==0])))))
    value = px.classification.f1v(X,y)
    assert value==1

def test_F1v_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.f1v(X,y)
    assert 0 <= value <= 1 
    

def test_F1v_one_class():
    X, y = make_classification(random_state=115, weights=[1, 0], flip_y=0)
    value = px.classification.f1v(X,y)
    assert np.isnan(value)

### F2 tests ###

def test_F2():
    metric = px.classification.f2
    assert _get_comparison(metric)>0.5

def test_F2_equal_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.f2(X,y)
    assert value==0

def test_F2_equal_to_one():
    X, y = make_classification(random_state=115)
    X = np.concatenate((X[y==0],X[y==0]), axis=0)
    y = np.concatenate((y[y==0],np.ones((len(y[y==0])))))
    value = px.classification.f2(X,y)
    assert value==1

def test_F2_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.f2(X,y)
    assert 0 <= value <= 1 

### F3 tests ###

def test_F3():
    metric = px.classification.f3
    assert _get_comparison(metric)>0.5

def test_F3_equal_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.f3(X,y)
    assert value==0

def test_F3_equal_to_one():
    X, y = make_classification(random_state=115)
    X = np.concatenate((X[y==0],X[y==0]), axis=0)
    y = np.concatenate((y[y==0],np.ones((len(y[y==0])))))
    value = px.classification.f3(X,y)
    assert value==1

def test_F3_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.f3(X,y)
    assert 0 <= value <= 1 

### F4 tests ###

def test_F4():
    metric = px.classification.f4
    assert _get_comparison(metric)>0.5

def test_F4_equal_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.f4(X,y)
    assert value==0

def test_F4_equal_to_one():
    X, y = make_classification(random_state=115)
    X = np.concatenate((X[y==0],X[y==0]), axis=0)
    y = np.concatenate((y[y==0],np.ones((len(y[y==0])))))
    value = px.classification.f4(X,y)
    assert value==1

def test_F4_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.f4(X,y)
    assert 0 <= value <= 1 

### Test L1 ###

def test_L1():
    metric = px.classification.l1
    assert _get_comparison(metric)>0.5

def test_L1_equal_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.l1(X,y)
    assert value==0

def test_L1_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.l1(X,y)
    assert 0 <= value <= 1 

### Test L2 ###

def test_L2():
    metric = px.classification.l2
    assert _get_comparison(metric)>0.5

def test_L2_equal_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.l2(X,y)
    assert value==0

def test_L2_random_classifier_accuracy():
    X, y = make_classification(random_state=115)
    X = np.concatenate((X[y==0],X[y==0]), axis=0)
    y = np.concatenate((y[y==0],np.ones((len(y[y==0])))))
    value = px.classification.l2(X,y)
    assert value==0.5

def test_L2_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.l2(X,y)
    assert 0 <= value <= 1 

### Test L3 ###

def test_L3():
    metric = px.classification.l3
    assert _get_comparison(metric)>0.5

def test_L3_equal_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.l3(X,y)
    assert value==0

def test_L3_random_classifier_accuracy():
    X, y = make_classification(random_state=115)
    X = np.concatenate((X[y==0],X[y==0]), axis=0)
    y = np.concatenate((y[y==0],np.ones((len(y[y==0])))))
    value = px.classification.l3(X,y)
    assert 0.4 <= value <= 0.6

def test_L3_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.l3(X,y)
    assert 0 <= value <= 1

### Test N1 ###

def test_N1():
    metric = px.classification.n1
    assert _get_comparison(metric)>0.5

def test_N1_close_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.n1(X,y)
    assert value<0.01

def test_N1_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.n1(X,y)
    assert 0 <= value <= 1 

### Test N2 ###

def test_N2():
    metric = px.classification.n2
    assert _get_comparison(metric)>0.5

def test_N2_close_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 1000
    value = px.classification.n2(X,y)
    assert value<0.01

def test_N2_equal_to_half():
    X, y = make_classification(random_state=115)
    X = np.concatenate((X[y==0],X[y==0]), axis=0)
    y = np.concatenate((y[y==0],np.ones((len(y[y==0])))))
    value = px.classification.n2(X,y)
    assert value==0.5

def test_N2_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.n2(X,y)
    assert 0 <= value <= 1 

def test_N2_single_class():
    X, y = make_classification(random_state=115, weights=[1., .0], flip_y=0.)
    value = px.classification.n2(X,y)
    assert np.isnan(value)

def test_N2_emptyset():
    X, y = [], []
    value = px.classification.n2(X,y)
    assert np.isnan(value)

### Test N3 ###

def test_N3():
    metric = px.classification.n3
    assert _get_comparison(metric)>0.5

def test_N3_equal_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.n3(X,y)
    assert value==0

def test_N3_equal_to_one():
    X, y = make_classification(random_state=115)
    X = np.concatenate((X[y==0],X[y==0]), axis=0)
    y = np.concatenate((y[y==0],np.ones((len(y[y==0])))))
    value = px.classification.n3(X,y)
    assert value==1

def test_N3_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.n3(X,y)
    assert 0 <= value <= 1

### Test N4 ###

def test_N4():
    metric = px.classification.n4
    assert _get_comparison(metric)>0.5

def test_N4_equal_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.n4(X,y)
    assert value==0

def test_N4_random_classifier_accuracy():
    X, y = make_classification(random_state=115)
    X = np.concatenate((X[y==0],X[y==0]), axis=0)
    y = np.concatenate((y[y==0],np.ones((len(y[y==0])))))
    value = px.classification.n4(X,y)
    assert 0.4 <= value <= 0.6

def test_N4_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.n4(X,y)
    assert 0 <= value <= 1

### Test T1 ###

def test_T1():
    metric = px.classification.t1
    assert _get_comparison(metric)>0.5

def test_T1_close_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.t1(X,y)
    assert value<=0.05

def test_T1_equal_to_one():
    X, y = make_classification(random_state=115)
    X = np.concatenate((X[y==0],X[y==0]), axis=0)
    y = np.concatenate((y[y==0],np.ones((len(y[y==0])))))    
    value = px.classification.t1(X,y)
    assert value==1

def test_T1_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.t1(X,y)
    assert 0 <= value <= 1

### Test LSC ###

def test_LSC():
    metric = px.classification.lsc
    assert _get_comparison(metric)>0.5

def test_LSC_close_to_zero():
    # depends on imbalance -- for balanced problems minimum is 0.5
    X, y = make_classification(random_state=115, weights=[0.99, 0.01])
    X[y==0] += 100
    value = px.classification.lsc(X,y)
    assert value<0.05

def test_LSC_equal_to_half():
    # depends on imbalance -- for balanced problems minimum is 0.5
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.lsc(X,y)
    assert value==0.5

def test_LSC_equal_to_one():
    X, y = make_classification(random_state=115)
    X = np.concatenate((X[y==0],X[y==0]), axis=0)
    y = np.concatenate((y[y==0],np.ones((len(y[y==0])))))
    value = px.classification.lsc(X,y)
    assert value==1

def test_LSC_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.lsc(X,y)
    assert 0 <= value <= 1

### Test Density ###

def test_density():
    metric = px.classification.density
    assert _get_comparison(metric)>0.5

def test_density_close_to_zero():
    # depends on imbalance -- for balanced problems minimum is 0.5
    X, y = make_classification(random_state=115, weights=[0.99, 0.01])
    X[y==0] += 100
    value = px.classification.density(X,y)
    assert value<0.05

def test_density_close_to_half():
    # depends on imbalance -- for balanced problems minimum is 0.5
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.density(X,y)
    assert 0.4 <= value <= 0.6

def test_density_equal_to_one():
    X, y = make_classification(random_state=115)
    X = np.concatenate((X[y==0],X[y==0]), axis=0)
    y = np.concatenate((y[y==0],np.ones((len(y[y==0])))))
    value = px.classification.density(X,y)
    assert value==1

def test_density_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.density(X,y)
    assert 0 <= value <= 1

### Test ClsCoef ###

def test_clsCoef():
    metric = px.classification.clsCoef
    assert _get_comparison(metric)>0.5

def test_clsCoef_equal_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.clsCoef(X,y)
    assert value==0

def test_clsCoef_equal_to_one():
    X, y = make_classification(random_state=115)
    X = np.concatenate((X[y==0],X[y==0]), axis=0)
    y = np.concatenate((y[y==0],np.ones((len(y[y==0])))))
    value = px.classification.clsCoef(X,y)
    assert value==1

def test_clsCoef_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.clsCoef(X,y)
    assert 0 <= value <= 1

### Test Hubs ###

def test_hubs():
    metric = px.classification.hubs
    assert _get_comparison(metric)>0.5

def test_hubs_equal_to_zero():
    X, y = make_classification(random_state=115)
    X[y==0] += 100
    value = px.classification.hubs(X,y)
    assert value==0

def test_hubs_close_to_one():
    X, y = make_classification(random_state=115, flip_y=0.5)
    value = px.classification.hubs(X,y)
    assert value>=0.95

def test_hubs_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.hubs(X,y)
    assert 0 <= value <= 1

### Test T2 ###

def test_T2():
    metric = px.classification.t2
    assert _get_comparison(metric)>0.5

def test_T2_close_to_zero():
    X, y = make_classification(random_state=115, n_features=2, n_informative=2, n_redundant=0, n_samples=1000)
    value = px.classification.t2(X,y)
    assert value==0.002

def test_T2_equal_to_one():
    X, y = make_classification(random_state=115, n_samples=20)
    value = px.classification.t2(X,y)
    assert value==1

def test_T2_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.t2(X,y)
    assert 0 <= value <= 1

### Test T3 ###

def test_T3():
    metric = px.classification.t3
    assert _get_comparison(metric)>0.5

def test_T3_close_to_zero():
    X, y = make_classification(random_state=115, n_features=20, n_informative=2, n_redundant=0, n_repeated=18, n_samples=1000)
    value = px.classification.t3(X,y)
    assert value==0.002

def test_T3_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.t3(X,y)
    assert 0 <= value <= 1

### Test T4 ###

def test_T4():
    metric = px.classification.t4
    assert _get_comparison(metric)>0.5

def test_T4_close_to_zero():
    X, y = make_classification(random_state=115, n_features=100, n_informative=2, n_redundant=0, n_repeated=98)
    value = px.classification.t4(X,y)
    assert value==0.02

def test_T4_imbalance():
    X, y = make_classification(random_state=115, weights=[.99, .01], flip_y=0.)
    value = px.classification.t4(X,y)
    assert 0 <= value <= 1

### Test C1 ###

def test_C1():
    metric = px.classification.c1
    assert _get_comparison(metric)>0.5

def test_C1_equal_to_zero():
    X, y = make_classification(random_state=115)
    value = px.classification.c1(X,y)
    assert value==0

def test_C1_close_to_one():
    X, y = make_classification(random_state=115, weights=[.999, .001], flip_y=0.)
    value = px.classification.c1(X,y)
    assert value>=0.9

### Test C2 ###

def test_C2():
    metric = px.classification.c2
    assert _get_comparison(metric)>0.5

def test_C2_equal_to_zero():
    X, y = make_classification(random_state=115)
    value = px.classification.c2(X,y)
    assert value==0

def test_C2_close_to_one():
    X, y = make_classification(random_state=115, weights=[.999, .001], flip_y=0.)
    value = px.classification.c2(X,y)
    assert value>=0.95

### Test ComplexityCalculator ###

def test_ComplexityCalculator():
    c = px.ComplexityCalculator()
    reps = 10
    res = np.zeros((reps, 2))
    rs = np.random.randint(100,10000,reps)    

    for r in range(reps):
        X_simple, y_simple = make_classification(n_samples=500, n_features=5, n_redundant=0, 
                        n_informative=5, n_clusters_per_class=2, n_classes=2, flip_y=0, 
                        class_sep=3.5, weights=[0.5, 0.5], random_state=rs[r])

        X_complex, y_complex = make_classification(n_samples=500, n_features=10, n_redundant=0, 
                        n_informative=10, n_clusters_per_class=2, n_classes=2, flip_y=0.2, 
                        class_sep=0.01, weights=[0.9, 0.1], random_state=rs[r])

        res[r, 0] = c.fit(X_simple, y_simple).score()
        res[r, 1] = c.fit(X_complex, y_complex).score()

    comparison = np.mean(res, axis=0)
    assert comparison[0]<comparison[1]
    
def test_ComplexityCalculator_report():
    c = px.ComplexityCalculator()

    X_simple, y_simple = make_classification(n_samples=500, n_features=5, n_redundant=0, 
                        n_informative=5, n_clusters_per_class=2, n_classes=2, flip_y=0, 
                        class_sep=3.5, weights=[0.5, 0.5], random_state=123)

    report = c.fit(X_simple, y_simple).report()
    assert isinstance(report, dict)
    
def test_OVA():
    c = px.ComplexityCalculator(multiclass_strategy='ova')
    reps = 10
    res = np.zeros((reps, 2))
    rs = np.random.randint(100,10000,reps)

    for r in range(reps):
        X_simple, y_simple = make_classification(n_samples=500, n_features=5, n_redundant=0, 
                        n_informative=5, n_clusters_per_class=2, n_classes=3, flip_y=0, 
                        class_sep=3.5, weights=[0.33, 0.33, 0.34], random_state=rs[r])

        X_complex, y_complex = make_classification(n_samples=500, n_features=10, n_redundant=0, 
                        n_informative=10, n_clusters_per_class=2, n_classes=3, flip_y=0.2, 
                        class_sep=0.01, weights=[0.8, 0.1, 0.1], random_state=rs[r])

        res[r, 0] = c.fit(X_simple, y_simple).score()
        res[r, 1] = c.fit(X_complex, y_complex).score()

    comparison = np.mean(res, axis=0)
    assert comparison[0]<comparison[1]
    
def test_OVO():
    c = px.ComplexityCalculator(multiclass_strategy='ovo')
    reps = 10
    res = np.zeros((reps, 2))
    rs = np.random.randint(100,10000,reps)

    for r in range(reps):
        X_simple, y_simple = make_classification(n_samples=500, n_features=5, n_redundant=0, 
                        n_informative=5, n_clusters_per_class=2, n_classes=3, flip_y=0, 
                        class_sep=3.5, weights=[0.33, 0.33, 0.34], random_state=rs[r])

        X_complex, y_complex = make_classification(n_samples=500, n_features=10, n_redundant=0, 
                        n_informative=10, n_clusters_per_class=2, n_classes=3, flip_y=0.2, 
                        class_sep=0.01, weights=[0.8, 0.1, 0.1], random_state=rs[r])

        res[r, 0] = c.fit(X_simple, y_simple).score()
        res[r, 1] = c.fit(X_complex, y_complex).score()

    comparison = np.mean(res, axis=0)
    assert comparison[0]<comparison[1]


"""
Smoothness measures
"""
import numpy as np
from scipy.sparse.csgraph import minimum_spanning_tree
from scipy.sparse import csr_matrix
from scipy.spatial import distance_matrix
from sklearn.model_selection import LeaveOneOut
from sklearn.neighbors import KNeighborsRegressor

def s1(X, y, normalize=True):
    """
    Calculates the output distribution (S1) measure.

    Calculates complexity based on a similarity of instances adjacent in minimum spanning tree (MST). Returns the average difference of labels (y), of samples connected by MST. By default a 0-1 interval normalization is performed.

    .. math::

        S1=\\frac{1}{n}\sum_{i,j \in MST}|y_i - y_j|

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels

    :rtype: float
    :returns: S1 score
    """

    X = np.copy(X)
    y = np.copy(y)

    if normalize:
        for f_id in range(X.shape[1]):
            X[:,f_id] -= np.min(X[:,f_id])
            X[:,f_id] /= np.max(X[:,f_id])
        y -= np.min(y)
        y /=np.max(y)
    
    dist = distance_matrix(X, X)
    graph = csr_matrix(dist)
    mst = minimum_spanning_tree(graph)
    mst = mst.toarray()
    coordinates = np.argwhere(mst>0)

    diff = np.abs(y[coordinates][:,0] - y[coordinates][:,1])

    return np.sum(diff)/X.shape[0]

def s2(X, y, normalize=True):
    """
    Calculates the input distribution (S2) measure.

    Calculates complexity based on a similarity of features (X) of instances with close output values (y). Returns the average euclidean norm of difference of input values, of samples neighbouring after sorting them by output values. By default a 0-1 interval normalization is performed.

    .. math::

        S2=\\frac{1}{n}\sum_{i=2}^{n}||x_i-x_{i-1}||_2

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels

    :rtype: float
    :returns: S2 score
    """

    X = np.copy(X)
    y = np.copy(y)

    if normalize:
        for f_id in range(X.shape[1]):
            X[:,f_id] -= np.min(X[:,f_id])
            X[:,f_id] /= np.max(X[:,f_id])
        y -= np.min(y)
        y /=np.max(y)
    
    sorted_y_idx = np.argsort(y)

    a = X[sorted_y_idx[:-1]]
    b = X[sorted_y_idx[1:]]
    diff = np.linalg.norm(a-b, axis=1)

    return np.sum(diff)/X.shape[0]

def s3(X, y, normalize=True):
    """
    Calculates the error of nearest neighbor regressor (S3) measure.

    Returns mean squared error of a 1-nearest neighbor regressor, established during leave-one-out procedure. By default, the data in normalized with 0-1 interval normalization.

    .. math::

        S3=\\frac{1}{n}\sum_{i=1}^{n}(NN(x_i)-y_i)^2

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels

    :rtype: float
    :returns: S3 score
    """
    X = np.copy(X)
    y = np.copy(y)

    if normalize:
        for f_id in range(X.shape[1]):
            X[:,f_id] -= np.min(X[:,f_id])
            X[:,f_id] /= np.max(X[:,f_id])
        y -= np.min(y)
        y /=np.max(y)

    loo = LeaveOneOut()

    err = []
    for train_index, test_index in loo.split(X):
        err.append(KNeighborsRegressor(n_neighbors=1).fit(X[train_index], y[train_index]).predict(X[test_index]) - y[test_index])

    return np.sum(np.power(err,2))/X.shape[0]


#! /usr/bin/env python
"""Toolbox for problem complexity."""
from __future__ import absolute_import

import os
# read the contents of your README file
from os import path

from setuptools import find_packages, setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# get __version__ from _version.py
ver_file = os.path.join("problexity", "_version.py")
with open(ver_file) as f:
    exec(f.read())

DISTNAME = "problexity"
DESCRIPTION = "The problexity module is an open-source python library containing the implementation of measures describing the complexity of the classification problem."
MAINTAINER = "J. Komorniczak"
MAINTAINER_EMAIL = "joanna.komorniczak@vp.pl"
URL = "https://w4k2.github.io/problexity/"
LICENSE = "GPL-3.0"
DOWNLOAD_URL = "https://github.com/w4k2/problexity"
VERSION = __version__
INSTALL_REQUIRES = ["numpy", "scipy", "scikit-learn", "matplotlib", "igraph", "pandas"]
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.7",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Topic :: Software Development :: Libraries",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

setup(
    name=DISTNAME,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    description=DESCRIPTION,
    license=LICENSE,
    url=URL,
    version=VERSION,
    download_url=DOWNLOAD_URL,
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=CLASSIFIERS,
)



"""
https://github.com/wwwjk366/gower

MIT License
Copyright (c) 2019 Michael Yan
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from scipy.sparse import issparse
import numpy as np
import pandas as pd

def gower_matrix(data_x, data_y=None, weight=None, cat_features=None):  
    
    # function checks
    X = data_x
    if data_y is None: Y = data_x 
    else: Y = data_y 
    if not isinstance(X, np.ndarray): 
        if not np.array_equal(X.columns, Y.columns): raise TypeError("X and Y must have same columns!")   
    else: 
         if not X.shape[1] == Y.shape[1]: raise TypeError("X and Y must have same y-dim!")  
                
    if issparse(X) or issparse(Y): raise TypeError("Sparse matrices are not supported!")

    # convert ordered categorical data to numerical (scaled values in interval 0->1)
    if isinstance(X, pd.DataFrame):
        X = X.copy()
        for column_name,column in X.iteritems():
            if isinstance(column.dtype, pd.CategoricalDtype) and column.dtype.ordered:
                X[column_name] = column.cat.codes.astype('float64') / (len(column.cat.categories) - 1)
        Y = X
            
    x_n_rows, x_n_cols = X.shape
    y_n_rows, y_n_cols = Y.shape 
    
    if cat_features is None:
        if not isinstance(X, np.ndarray): 
            is_number = np.vectorize(lambda x: not np.issubdtype(x, np.number))
            cat_features = is_number(X.dtypes)    
        else:
            cat_features = np.zeros(x_n_cols, dtype=bool)
            for col in range(x_n_cols):
                if not np.issubdtype(type(X[0, col]), np.number):
                    cat_features[col]=True
    else:          
        cat_features = np.array(cat_features)
    
    # print(cat_features)
    
    if not isinstance(X, np.ndarray): X = np.asarray(X)
    if not isinstance(Y, np.ndarray): Y = np.asarray(Y)
    
    Z = np.concatenate((X,Y))
    
    x_index = range(0,x_n_rows)
    y_index = range(x_n_rows,x_n_rows+y_n_rows)
    
    Z_num = Z[:,np.logical_not(cat_features)]
    
    num_cols = Z_num.shape[1]
    num_ranges = np.zeros(num_cols)
    num_max = np.zeros(num_cols)
    
    for col in range(num_cols):
        col_array = Z_num[:, col].astype(np.float32) 
        max = np.nanmax(col_array)
        min = np.nanmin(col_array)
     
        if np.isnan(max):
            max = 0.0
        if np.isnan(min):
            min = 0.0
        num_max[col] = max
        num_ranges[col] = (1 - min / max) if (max != 0) else 0.0

    # This is to normalize the numeric values between 0 and 1.
    Z_num = np.divide(Z_num ,num_max,out=np.zeros_like(Z_num), where=num_max!=0)
    Z_cat = Z[:,cat_features]
    
    if weight is None:
        weight = np.ones(Z.shape[1])
        
    #print(weight)    
    
    weight_cat=weight[cat_features]
    weight_num=weight[np.logical_not(cat_features)]   
        
    out = np.zeros((x_n_rows, y_n_rows), dtype=np.float32)
        
    weight_sum = weight.sum()
    
    X_cat = Z_cat[x_index,]
    X_num = Z_num[x_index,]
    Y_cat = Z_cat[y_index,]
    Y_num = Z_num[y_index,]
    
   # print(X_cat,X_num,Y_cat,Y_num)
    
    for i in range(x_n_rows):          
        j_start= i        
        if x_n_rows != y_n_rows:
            j_start = 0
        # call the main function
        res = gower_get(X_cat[i,:], 
                          X_num[i,:],
                          Y_cat[j_start:y_n_rows,:],
                          Y_num[j_start:y_n_rows,:],
                          weight_cat,
                          weight_num,
                          weight_sum,
                          cat_features,
                          num_ranges,
                          num_max) 
        #print(res)
        out[i,j_start:]=res
        if x_n_rows == y_n_rows: out[i:,j_start]=res
        
    return out


def gower_get(xi_cat,xi_num,xj_cat,xj_num,feature_weight_cat,
              feature_weight_num,feature_weight_sum,categorical_features,
              ranges_of_numeric,max_of_numeric ):
    
    # categorical columns
    sij_cat = np.where(xi_cat == xj_cat,np.zeros_like(xi_cat),np.ones_like(xi_cat))
    sum_cat = np.multiply(feature_weight_cat,sij_cat).sum(axis=1) 

    # numerical columns
    abs_delta=np.absolute(xi_num-xj_num)
    sij_num=np.divide(abs_delta, ranges_of_numeric, out=np.zeros_like(abs_delta), where=ranges_of_numeric!=0)

    sum_num = np.multiply(feature_weight_num,sij_num).sum(axis=1)
    sums= np.add(sum_cat,sum_num)
    sum_sij = np.divide(sums,feature_weight_sum)
    
    return sum_sij